# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

# AUTOGENERATED FILE - No need to edit
# GENERATED AT: {{ now().strftime('%Y-%m-%dT%H:%M:%S%z"') }}
# Re-create by running 'generate-routes' command
# Run 'generate-routes --help' for more options
{% block template %}
{% endblock %}
